/*
 * Copyright 2012, 2017 INFN
 *
 * Licensed under the EUPL, Version 1.2 or – as soon they
 * will be approved by the European Commission - subsequent
 * versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the
 * Licence.
 * You may obtain a copy of the Licence at:
 *
 * https://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in
 * writing, software distributed under the Licence is
 * distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied.
 * See the Licence for the specific language governing
 * permissions and limitations under the Licence.
 */

#include <chaos/common/network/NetworkBroker.h>
#include <chaos/common/configuration/GlobalConfiguration.h>
#include <chaos/common/metadata_logging/ErrorLoggingChannel.h>
#include <chaos/common/metadata_logging/AlarmLoggingChannel.h>
#include <chaos/common/metadata_logging/StandardLoggingChannel.h>
#include <chaos/common/metadata_logging/MetadataLoggingManager.h>
#include <chaos/common/metadata_logging/BatchCommandLoggingChannel.h>

using namespace chaos;
using namespace chaos::common::data;
using namespace chaos::common::pqueue;
using namespace chaos::common::network;
using namespace chaos::common::message;
using namespace chaos::common::metadata_logging;

#define MLM_INFO    INFO_LOG(MetadataLoggingManager)
#define MLM_DBG     DBG_LOG(MetadataLoggingManager)
#define MLM_ERR     ERR_LOG(MetadataLoggingManager)

#ifdef USE_MESSAGE_PUBLISH_SUBCRIBE
MetadataLoggingManager::MetadataLoggingManager() {
    //add default channels
    registerChannel("StandardLoggingChannel",
                    METADATA_LOGGING_STANDARD_LOGGING_CHANNEL_INSTANCER);
    registerChannel("AlarmLoggingChannel",
                    METADATA_LOGGING_ALARM_LOGGING_CHANNEL_INSTANCER);
    registerChannel("ErrorLoggingChannel",
                    METADATA_LOGGING_ERROR_LOGGING_CHANNEL_INSTANCER);
    registerChannel("BatchCommandLoggingChannel",
                    METADATA_LOGGING_COMMAND_LOGGING_CHANNEL_INSTANCER);
}

MetadataLoggingManager::~MetadataLoggingManager() {}

void MetadataLoggingManager::init(void *init_data)  {
    std::string msgbrokerdrv = "kafka-rdk";
    msgbrokerdrv = GlobalConfiguration::getInstance()->getOption<std::string>(InitOption::OPT_MSG_BROKER_DRIVER);

    prod            = chaos::common::message::MessagePSDriver::getProducerDriver(msgbrokerdrv);
  
    CHAOS_LASSERT_EXCEPTION(prod, MLM_ERR, -1 , "Error getting producer message");
    
    
}

void MetadataLoggingManager::deinit()  {
    
    ChaosLockGuard wl(mutext_maps);
    for(MetadataLoggingInstancesMapIterator it = map_instance.begin();
        it != map_instance.end();
        it++) {
        MLM_DBG << "Remove channel instance:"<< it->first;
        if(it->second != NULL) {
            delete(it->second);
        }
    }
    MLM_DBG << "All channel has been removed";
    map_instance.clear();
    map_instancer.clear();
    
}

void MetadataLoggingManager::registerChannel(const std::string& channel_alias,
                                             chaos::common::utility::ObjectInstancer<AbstractMetadataLogChannel> *instancer) {
    map_instancer.insert(std::pair<std::string, ChaosSharedPtr< chaos::common::utility::ObjectInstancer<AbstractMetadataLogChannel > > >(channel_alias, ChaosSharedPtr< chaos::common::utility::ObjectInstancer<AbstractMetadataLogChannel > > (instancer)));
}

AbstractMetadataLogChannel *MetadataLoggingManager::getChannel(const std::string channel_alias) {
    //chec if we are initilized
    if(getServiceState() != 1) return NULL;
    ChaosLockGuard wl(mutext_maps);
    //check if we have the channel
    if(map_instancer.count(channel_alias) == 0) {
        MLM_ERR<<" No channels registered: "+channel_alias;
        return NULL;
    }
    
    AbstractMetadataLogChannel *result = map_instancer[channel_alias]->getInstance();
    if(result==NULL){
         MLM_ERR<<" Invalid channel:"+channel_alias;
        return NULL;
    }
    result->setLoggingManager(this);
    map_instance.insert(std::pair<std::string, AbstractMetadataLogChannel*> (result->getInstanceUUID(), result));
    
    MLM_DBG << "Created new channel instance " << result->getInstanceUUID() << " for " << channel_alias;
    return result;
}

void MetadataLoggingManager::releaseChannel(AbstractMetadataLogChannel *channel_instance) {
    //chec if we are initilized
    ChaosLockGuard wl(mutext_maps);
    if(channel_instance == NULL) return;
    if(map_instance.count(channel_instance->getInstanceUUID()) == 0) return;
    //we can delete the instance
    
    map_instance.erase(channel_instance->getInstanceUUID());
    MLM_DBG << "Release channel instance " << channel_instance->getInstanceUUID();
    delete(channel_instance);
}
 int  MetadataLoggingManager::pushLogEntry(chaos::common::data::CDataWrapper *log_entry,
                                 int32_t priority ){
                                     int err=0;
if(log_entry&& log_entry->hasKey(MetadataServerLoggingDefinitionKeyRPC::PARAM_NODE_LOGGING_LOG_SOURCE_IDENTIFIER)){
std::string key=log_entry->getStringValue(MetadataServerLoggingDefinitionKeyRPC::PARAM_NODE_LOGGING_LOG_SOURCE_IDENTIFIER);
//std::string key="CHAOS_LOG";
log_entry->addInt32Value(DataPackCommonKey::DPCK_DATASET_TYPE,DataPackCommonKey::DPCK_DATASET_TYPE_LOG);
log_entry->addStringValue(NodeDefinitionKey::NODE_UNIQUE_ID,key);

//MLM_DBG <<key<< ": LOGGING " <<log_entry->getJSONString();

if ((err = prod->pushMsgAsync(*log_entry, key)) != 0) {
    MLM_ERR << "Error pushing log entry" << prod->getLastError();
  }
}
  
return 0;
}


#else
MetadataLoggingManager::MetadataLoggingManager():
message_channel(NULL) {
    //add default channels
    registerChannel("StandardLoggingChannel",
                    METADATA_LOGGING_STANDARD_LOGGING_CHANNEL_INSTANCER);
    registerChannel("AlarmLoggingChannel",
                    METADATA_LOGGING_ALARM_LOGGING_CHANNEL_INSTANCER);
    registerChannel("ErrorLoggingChannel",
                    METADATA_LOGGING_ERROR_LOGGING_CHANNEL_INSTANCER);
    registerChannel("BatchCommandLoggingChannel",
                    METADATA_LOGGING_COMMAND_LOGGING_CHANNEL_INSTANCER);
}

MetadataLoggingManager::~MetadataLoggingManager() {}

void MetadataLoggingManager::init(void *init_data)  {
    message_channel = NetworkBroker::getInstance()->getMetadataserverMessageChannel();
    CHAOS_LASSERT_EXCEPTION(message_channel, MLM_ERR, -1 , "Error gettin mds channel");
    
    CObjectProcessingPriorityQueue<CDataWrapper>::init(1);
}

void MetadataLoggingManager::deinit()  {
    MLM_DBG << "Wait for queue will empty";
    CObjectProcessingPriorityQueue<CDataWrapper>::deinit(true);
    MLM_DBG << "Queue is empty";
    
    ChaosLockGuard wl(mutext_maps);
    for(MetadataLoggingInstancesMapIterator it = map_instance.begin();
        it != map_instance.end();
        it++) {
        MLM_DBG << "Remove channel instance:"<< it->first;
        if(it->second != NULL) {
            delete(it->second);
        }
    }
    MLM_DBG << "All channel has been removed";
    map_instance.clear();
    map_instancer.clear();
    
    if(message_channel != NULL){NetworkBroker::getInstance()->disposeMessageChannel(message_channel); message_channel = NULL;}
}

void MetadataLoggingManager::registerChannel(const std::string& channel_alias,
                                             chaos::common::utility::ObjectInstancer<AbstractMetadataLogChannel> *instancer) {
    map_instancer.insert(std::pair<std::string, ChaosSharedPtr< chaos::common::utility::ObjectInstancer<AbstractMetadataLogChannel > > >(channel_alias, ChaosSharedPtr< chaos::common::utility::ObjectInstancer<AbstractMetadataLogChannel > > (instancer)));
}

AbstractMetadataLogChannel *MetadataLoggingManager::getChannel(const std::string channel_alias) {
    //chec if we are initilized
    if(getServiceState() != 1) return NULL;
    ChaosLockGuard wl(mutext_maps);
    //check if we have the channel
    if(map_instancer.count(channel_alias) == 0) return NULL;
    
    AbstractMetadataLogChannel *result = map_instancer[channel_alias]->getInstance();
    result->setLoggingManager(this);
    map_instance.insert(std::pair<std::string, AbstractMetadataLogChannel*> (result->getInstanceUUID(), result));
    
    MLM_DBG << "Created new channel instance " << result->getInstanceUUID() << " for " << channel_alias;
    return result;
}

void MetadataLoggingManager::releaseChannel(AbstractMetadataLogChannel *channel_instance) {
    //chec if we are initilized
    ChaosLockGuard wl(mutext_maps);
    if(channel_instance == NULL) return;
    if(map_instance.count(channel_instance->getInstanceUUID()) == 0) return;
    //we can delete the instance
    
    map_instance.erase(channel_instance->getInstanceUUID());
    MLM_DBG << "Release channel instance " << channel_instance->getInstanceUUID();
    delete(channel_instance);
}

void MetadataLoggingManager::processBufferElement(CDWShrdPtr log_entry) {
    CHAOS_ASSERT(getServiceState() == 1);
    int err = 0;
    DEBUG_CODE(MLM_DBG << "forwarding log entry " << log_entry->getJSONString());
    
    //detach the entry from the queue
    if(message_channel) {
        if((err = sendLogEntry(MOVE(log_entry->clone())))) {
            MLM_ERR << "Error forwarding log entry with code:" << err;
            //log entry need to be resubmitted or stored on disk (in future version)
            //delete(log_entry);
        } else {
            //log entry has been submitted
            MLM_DBG << "Successfully submited log entry";
        }
    } else {
        //no message channel available so we need to store the log
        MLM_DBG << "No channel for submitting log entry";
    }
}

int MetadataLoggingManager::sendLogEntry(CDWUniquePtr log_entry) {
    CHAOS_ASSERT(getServiceState() == 1);
    
    int err = 0;
    //send message to mds and wait for ack
    ChaosUniquePtr<MultiAddressMessageRequestFuture> log_future = message_channel->sendRequestWithFuture(MetadataServerLoggingDefinitionKeyRPC::ACTION_NODE_LOGGING_RPC_DOMAIN,
                                                                                                         MetadataServerLoggingDefinitionKeyRPC::ACTION_NODE_LOGGING_SUBMIT_ENTRY,
                                                                                                         MOVE(log_entry),
                                                                                                         2000);
    //wait for ack
    if(log_future.get()&&log_future->wait()) {
        err = log_future->getError();
    } else {
        //we can't be able to send log to any mds server so detach it and need to store it
        DEBUG_CODE(MLM_ERR << "we can't be able to send log to any mds server");
        err = -10000;
    }
    return err;
}

int MetadataLoggingManager::pushLogEntry(chaos::common::data::CDataWrapper *log_entry,
                                         int32_t priority) {
    CHAOS_ASSERT(getServiceState() == 1);
    if(log_entry == NULL) return -10000;
    PriorityQueuedElement<CDataWrapper>::PriorityQueuedElementType element(log_entry);
    return CObjectProcessingPriorityQueue<CDataWrapper>::push(element,
                                                              priority);
}
#endif
